import unittest
from unittest import TestCase
from unittest.mock import MagicMock, patch

from jwt import JWT
from oic.oic.consumer import Consumer
from oic.utils.authn.client import CLIENT_AUTHN_METHOD
from oic.utils.session_backend import DictSessionBackend


class Issue853TestCase(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        # Persistent session backend
        cls.session_backend = DictSessionBackend()

        # real JWT payload from a captured keycloak backchannel-logout request
        cls.body = "eyJhbGciOiJSUzI1NiIsInR5cCIgOiAiSldUIiwia2lkIiA6ICJkRVl0UnRmVHpBbXpNZGxlWU5aSWhvQnI3bFNieHBQU3Z4N3Y1eWVCUlU4In0.eyJpYXQiOjE2ODAwOTA5NDUsImp0aSI6ImIzMGFjOTEyLWRhYjctNDkzZC04MGQxLTM2NWU0NmEwOWUyNyIsImlzcyI6Imh0dHA6Ly9rZXljbG9hay5sb2NhbDo4MDgwL2F1dGgvcmVhbG1zL0RlbW8iLCJhdWQiOiJmdWxsIiwic3ViIjoiYWJlYzcwZjgtYTJhMC00YmYwLTg2NjUtYWI5Mzg3YTRkZTAyIiwidHlwIjoiTG9nb3V0Iiwic2lkIjoiNDRlNTliOTYtYTVhMy00NWEwLWI0ZWYtNzY3NzdmNGI3YzI1IiwiZXZlbnRzIjp7Imh0dHA6Ly9zY2hlbWFzLm9wZW5pZC5uZXQvZXZlbnQvYmFja2NoYW5uZWwtbG9nb3V0Ijp7fX19.D4XFSfVp8_T4TwDWLvHx--rs9-aLS8ZbZNPMNWIanoil9gc3N8UczsHJqqTQVQU7BNDQKeMVZQn47I3A1gW9_5WhQa5Si5xgKUmBTs-BoUjLo9Avr7lqAc7zOcGD4ehVLX6gv3PxAlD04snqjEBBW2PZYOZ04u0E--Ssbqd_LAha7ArbgMDG8dIBmJUHvJNMhWERX3QKw5cc3TXcY1TbZ-xDdkBf28DJ19ryXMHn0PybH927ZsDGX-2vxltDFRbIhotPXfoAbfZl8_TA84tn58zKcWKNd5aMtZ5Mu0D_SHPcNNpbYR3WzMlQZ6E0_8io5_buUBehaKZTryL30rK56w"

    @unittest.skip("This is a test for : https://github.com/CZ-NIC/pyoidc/issues/853")
    def test_bug(self):

        consumer_config = {
            # "debug": True,
            "response_type": "code",
        }

        client_config = {
            "client_id": "test",
            "client_authn_method": CLIENT_AUTHN_METHOD,
        }

        consumer = Consumer(
            consumer_config=consumer_config,
            client_config=client_config,
            session_db=self.session_backend,
        )

        with patch("oic.oic.consumer.BackChannelLogoutRequest") as mock:
            # Load jwt as python dict
            decoded = {"logout_token": JWT().decode(self.body, do_verify=False)}

            # Simulate previously made login : our session id is 0000
            real_sid = "0000"
            self.session_backend.update(
                real_sid, "smid", decoded["logout_token"]["sid"]
            )
            self.session_backend.update(real_sid, "sub", decoded["logout_token"]["sub"])
            self.session_backend.update(
                real_sid, "issuer", decoded["logout_token"]["iss"]
            )

            # Mock 'BackChannelLogoutRequest' as we don't want to check signature and stuff
            mocked_logout_request = MagicMock()
            mocked_logout_request.__getitem__.side_effect = decoded.__getitem__
            mock.return_value = mocked_logout_request

            # Uncomment the following line to have the test pass :
            # consumer.sso_db = consumer.sdb

            sid = consumer.backchannel_logout(request_args={"logout_token": self.body})

            self.assertIsNotNone(sid)
            self.assertEqual(sid, real_sid)



from importlib import import_module
from unittest import mock
from unittest.mock import ANY, call

from django.conf import settings
from django.contrib.auth import SESSION_KEY, get_user_model
from django.urls import reverse
from jwt import JWT, jwk_from_dict
from oic.oic import IdToken
from oic.oic.message import OpenIDSchema

from django_pyoidc.client import OIDCClient
from django_pyoidc.models import OIDCSession
from tests.utils import OIDCTestCase

SessionStore = import_module(settings.SESSION_ENGINE).SessionStore


class LoginViewTestCase(OIDCTestCase):
    @mock.patch("django_pyoidc.client.Consumer.provider_config")
    @mock.patch(
        "django_pyoidc.client.Consumer.begin",
        return_value=(1234, "https://sso.notatld"),
    )
    def test_redirect_uri_management_no_next_params(self, *args):
        """
        Test that without a next parameter we are redirected to 'post_logout_redirect_uri'
        """
        response = self.client.get(
            reverse("test_login"),
            SERVER_NAME="test.django-pyoidc.notatld",
        )
        self.assertRedirects(
            response, "https://sso.notatld", fetch_redirect_response=False
        )
        self.assertEqual(
            self.client.session["oidc_login_next"],
            settings.DJANGO_PYOIDC["sso1"]["post_login_uri_success"],
        )

    @mock.patch("django_pyoidc.client.Consumer.provider_config")
    @mock.patch(
        "django_pyoidc.client.Consumer.begin",
        return_value=(1234, "https://sso.notatld"),
    )
    def test_redirect_uri_management_next_to_samesite(self, *args):
        """
        Test that redirecting to a site allowed in 'login_uris_redirect_allowed_hosts' works
        """
        response = self.client.get(
            reverse("test_login"),
            data={
                "next": "https://"
                + settings.DJANGO_PYOIDC["sso1"]["login_uris_redirect_allowed_hosts"][0]
                + "/myview/details"
            },
            SERVER_NAME="test.django-pyoidc.notatld",
        )
        self.assertRedirects(
            response, "https://sso.notatld", fetch_redirect_response=False
        )
        self.assertEqual(
            self.client.session["oidc_login_next"],
            "https://test.django-pyoidc.notatld/myview/details",
        )

    @mock.patch("django_pyoidc.client.Consumer.provider_config")
    @mock.patch(
        "django_pyoidc.client.Consumer.begin",
        return_value=(1234, "https://sso.notatld"),
    )
    def test_redirect_uri_management_next_follows_https_requires(self, *args):
        """
        Test that trying to redirect to a non https site when login_redirection_requires_https is set to True does not work
        """
        response = self.client.get(
            reverse("test_login"),
            data={
                "next": "http://"
                + settings.DJANGO_PYOIDC["sso1"]["login_uris_redirect_allowed_hosts"][0]
                + "/myview/details"
            },
            SERVER_NAME="test.django-pyoidc.notatld",
        )
        self.assertRedirects(
            response, "https://sso.notatld", fetch_redirect_response=False
        )
        self.assertEqual(
            self.client.session["oidc_login_next"],
            settings.DJANGO_PYOIDC["sso1"]["post_login_uri_success"],
        )

    @mock.patch("django_pyoidc.client.Consumer.provider_config")
    def test_redirect_uri_bad_server_name(self, *args):
        """
        Test that requesting django oidc with bad host name is rejected (HTTP 400).
        """
        response = self.client.get(
            reverse("test_login"),
            data={"next": "test.django-pyoidc.notatld/myview/details"},
            SERVER_NAME="test.hacker.notatld",
        )
        self.assertEqual(response.status_code, 400)

    @mock.patch("django_pyoidc.client.Consumer.provider_config")
    def test_redirect_uri_management_next_to_disallowed_site(self, *args):
        """
        Test that trying to redirect to a site not allowed in 'login_uris_redirect_allowed_hosts' results in ignored instruction.

        The library will reject this host and use the 'post_login_uri_success' setting instead.
        """
        response = self.client.get(
            reverse("test_login"),
            data={"next": "http://test.hacker.notatld/myview/details"},
            SERVER_NAME="test.django-pyoidc.notatld",
        )
        self.assertEqual(response.status_code, 302)
        self.assertEqual(
            self.client.session["oidc_login_next"],
            settings.DJANGO_PYOIDC["sso1"]["post_login_uri_success"],
        )

    @mock.patch("django_pyoidc.client.Consumer.provider_config")
    def test_redirect_uri_management_next_to_disallowed_site2(self, *args):
        """
        Test that trying to redirect to a site not allowed in 'login_uris_redirect_allowed_hosts' results in ignored instruction.

        The library will reject this host and use  "/" instead (as 'post_login_uri_success' is undefined for sso2).
        """
        response = self.client.get(
            reverse("test_login_2"),
            data={"next": "http://test.hacker.notatld/myview/details"},
            SERVER_NAME="test2.django-pyoidc.notatld",
        )
        self.assertEqual(response.status_code, 302)
        self.assertEqual(
            self.client.session["oidc_login_next"],
            "/",
        )

    @mock.patch("django_pyoidc.client.Consumer.provider_config")
    def test_oidc_session_is_saved(self, *args):
        """
        Test that the OIDC client is saved on login request, and that the returned session ID allows us to restore the client.
        """
        response = self.client.get(
            reverse("test_login"),
            data={
                "next": "https://"
                + settings.DJANGO_PYOIDC["sso1"]["login_uris_redirect_allowed_hosts"][0]
                + "/myview/details"
            },
            SERVER_NAME="test.django-pyoidc.notatld",
        )
        self.assertEqual(response.status_code, 302)
        sid = self.client.session["oidc_sid"]
        self.assertIsNotNone(sid)
        client = OIDCClient(op_name="sso1", session_id=sid)
        self.assertEqual(client.consumer.client_id, "1")


class LogoutViewTestCase(OIDCTestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user = get_user_model().objects.create(
            username="test_user", email="test_user"
        )

    def test_logout_user_not_authenticated(self):
        """
        Test that trying to logout while not being connected redirects
        """
        response = self.client.get(reverse("test_logout"))
        self.assertRedirects(
            response, "http://testserver/logoutdone", fetch_redirect_response=False
        )

    def test_django_user_is_at_least_logged_out(self):
        """
        Test that logging out without any OIDC state at least kills the Django session
        """
        self.client.force_login(self.user)
        response = self.client.get(reverse("test_logout"))
        self.assertRedirects(
            response, "http://testserver/logoutdone", fetch_redirect_response=False
        )
        self.assertFalse(
            SESSION_KEY in self.client.session
        )  # from https://stackoverflow.com/a/6013115

    @mock.patch("django_pyoidc.client.Consumer.restore")
    @mock.patch(
        "django_pyoidc.client.Consumer.request_info",
        return_value=("http://example.com", "", "", ""),
    )
    def test_logout_generates_oidc_request_to_sso(
        self, mocked_request_info, mocked_restore
    ):
        """
        Test that logging out while being connected and having a valid OIDC session triggers an OIDC request to the SSO
        noticing it of the logout.
        """
        self.client.force_login(self.user)

        sid = "test_id_12345"

        session = self.client.session
        session["oidc_sid"] = sid
        session.save()

        response = self.client.get(reverse("test_logout"))
        self.assertRedirects(
            response, "http://example.com", fetch_redirect_response=False
        )
        self.assertFalse(
            SESSION_KEY in self.client.session
        )  # from https://stackoverflow.com/a/6013115
        mocked_request_info.assert_called_once()
        mocked_restore.assert_called_once_with(sid)


class CallbackViewTestCase(OIDCTestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user = get_user_model().objects.create(
            username="test_user", email="test_user"
        )

    def test_callback_but_no_sid_on_our_side(self):
        """
        Test that receiving a random request without any session states is well handled.
        """
        response = self.client.get(reverse("my_test_callback"))

        self.assertRedirects(response, "/logout_failure", fetch_redirect_response=False)

    @mock.patch(
        "django_pyoidc.client.Consumer.parse_authz",
        return_value=({"state": ""}, None, None),
    )
    @mock.patch("django_pyoidc.client.Consumer.restore")
    def test_callback_but_state_mismatch(self, mocked_restore, mocked_parse_authz):
        """
        Test that receiving a callback with a wrong state parameter results in an HTTP 4XX error.
        """
        self.client.force_login(self.user)

        state = "test_id_12345"

        session = self.client.session
        session["oidc_sid"] = state
        session.save()

        response = self.client.get(reverse("my_test_callback"))
        self.assertEqual(response.status_code, 400)
        mocked_restore.assert_called_once_with(state)
        mocked_parse_authz.assert_called_once()

    @mock.patch(
        "django_pyoidc.client.Consumer.parse_authz",
        return_value=({"state": "test_id_12345"}, None, None),
    )
    @mock.patch("django_pyoidc.engine.get_user_by_email", return_value=None)
    @mock.patch(
        "django_pyoidc.client.Consumer.get_user_info", return_value=OpenIDSchema()
    )
    @mock.patch(
        "django_pyoidc.client.Consumer.complete",
        return_value={"id_token": IdToken(iss="fake"), "access_token": "--"},
    )
    @mock.patch("django_pyoidc.client.Consumer.restore")
    def test_callback_no_session_state_provided_invalid_user(
        self,
        mocked_restore,
        mocked_complete,
        mocked_get_user_info,
        mocked_get_user,
        mocked_parse_authz,
    ):
        """
        Test that receiving a callback for a user that does not get validated by the developer-provided function 'get_user'
        does not get logged in
        """
        self.client.force_login(self.user)

        state = "test_id_12345"

        session = self.client.session
        session["oidc_sid"] = state
        session.save()

        response = self.client.get(reverse("my_test_callback"))
        mocked_restore.assert_called_once_with(state)
        mocked_complete.assert_called_once_with(state=state, session_state=None)
        mocked_parse_authz.assert_called_once()
        mocked_get_user_info.assert_called_once_with(state=state)
        mocked_get_user.assert_called_once_with(
            {
                "info_token_claims": {},
                "access_token_jwt": "--",
                "id_token_claims": {"iss": "fake"},
            }
        )

        self.assertRedirects(response, "/logout_failure", fetch_redirect_response=False)
        self.assertEqual(OIDCSession.objects.all().count(), 0)

    @mock.patch("django_pyoidc.views.OIDCView.call_user_login_callback_function")
    @mock.patch(
        "django_pyoidc.client.Consumer.parse_authz",
        return_value=({"state": "test_id_12345"}, None, None),
    )
    @mock.patch("django_pyoidc.engine.get_user_by_email")
    @mock.patch("django_pyoidc.client.Consumer.get_user_info")
    @mock.patch(
        "django_pyoidc.client.Consumer.complete",
        return_value={"id_token": IdToken(iss="fake"), "access_token": "--"},
    )
    @mock.patch("django_pyoidc.client.Consumer.restore")
    def test_callback_no_session_state_provided_valid_user(
        self,
        mocked_restore,
        mocked_complete,
        mocked_get_user_info,
        mocked_get_user,
        mocked_parse_authz,
        mocked_call_user_login_callback_function,
    ):
        """
        Test that receiving a callback for a user that gets validated by the developer-provided function 'get_user'
        is logged_in
        """
        self.client.force_login(self.user)

        state = "test_id_12345"

        session = self.client.session
        session["oidc_sid"] = state
        session.save()

        user_info = OpenIDSchema(sub="aaaaaeeee")
        user_info_dict = {"sub": "aaaaaeeee"}
        mocked_get_user_info.return_value = user_info

        dummy_user = self.user
        mocked_get_user.return_value = dummy_user

        response = self.client.get(reverse("my_test_callback"))

        with self.subTest("pyoidc calls are performed"):
            mocked_restore.assert_called_once_with(state)
            mocked_complete.assert_called_once_with(state=state, session_state=None)
            mocked_parse_authz.assert_called_once()
            mocked_get_user_info.assert_called_once_with(state=state)

        mocked_get_user.assert_called_once_with(
            {
                "info_token_claims": user_info_dict,
                "access_token_jwt": "--",
                "id_token_claims": {"iss": "fake"},
            }
        )

        self.assertRedirects(
            response, "/default/success", fetch_redirect_response=False
        )
        with self.subTest("Session is created correctly :"):
            self.assertEqual(OIDCSession.objects.all().count(), 1)

            session = OIDCSession.objects.all().first()

            self.assertEqual(session.session_state, None)
            self.assertEqual(session.sub, user_info_dict["sub"])
            self.assertEqual(session.state, state)
            self.assertEqual(session.cache_session_key, self.client.session.session_key)
        mocked_call_user_login_callback_function.assert_called_once()

    @mock.patch("django_pyoidc.views.OIDCView.call_user_login_callback_function")
    @mock.patch("django_pyoidc.client.Consumer.parse_authz")
    @mock.patch("django_pyoidc.engine.get_user_by_email")
    @mock.patch("django_pyoidc.client.Consumer.get_user_info")
    @mock.patch(
        "django_pyoidc.client.Consumer.complete",
        return_value={"id_token": IdToken(iss="fake"), "access_token": "--"},
    )
    @mock.patch("django_pyoidc.client.Consumer.restore")
    def test_callback_with_session_state_provided_valid_user(
        self,
        mocked_restore,
        mocked_complete,
        mocked_get_user_info,
        mocked_get_user,
        mocked_parse_authz,
        mocked_call_user_login_callback_function,
    ):
        """
        Test that receiving a callback with a session state (SID) for a user that gets validated by the developer-provided
        function 'get_user' is logged_in
        """
        self.client.force_login(self.user)

        state = "test_id_12345"
        session_state = "fe634"

        session = self.client.session
        session["oidc_sid"] = state
        session.save()

        user_info = OpenIDSchema(sub="aaaaaeeee")
        user_info_dict = {"sub": "aaaaaeeee"}
        mocked_get_user_info.return_value = user_info

        authz = {"state": state, "session_state": session_state}
        mocked_parse_authz.return_value = authz, None, None

        dummy_user = self.user
        mocked_get_user.return_value = dummy_user

        response = self.client.get(reverse("my_test_callback"))

        with self.subTest("pyoidc calls are performed"):
            mocked_restore.assert_called_once_with(state)
            mocked_complete.assert_called_once_with(
                state=state, session_state=session_state
            )
            mocked_parse_authz.assert_called_once()
            mocked_get_user_info.assert_called_once_with(state=state)

        mocked_get_user.assert_called_once_with(
            {
                "info_token_claims": user_info_dict,
                "access_token_jwt": "--",
                "id_token_claims": {"iss": "fake"},
            }
        )

        self.assertRedirects(
            response, "/default/success", fetch_redirect_response=False
        )

        with self.subTest("Session is created correctly :"):
            self.assertEqual(OIDCSession.objects.all().count(), 1)

            session = OIDCSession.objects.all().first()

            self.assertEqual(session.session_state, session_state)
            self.assertEqual(session.sub, user_info["sub"])
            self.assertEqual(session.state, state)
            self.assertEqual(session.cache_session_key, self.client.session.session_key)

        mocked_call_user_login_callback_function.assert_called_once()

    @mock.patch(
        "django_pyoidc.client.Consumer.parse_authz",
        return_value=({"state": "test_id_12345"}, None, None),
    )
    @mock.patch("django_pyoidc.engine.get_user_by_email", return_value=None)
    @mock.patch(
        "django_pyoidc.client.Consumer.get_user_info", return_value=OpenIDSchema()
    )
    @mock.patch(
        "django_pyoidc.client.Consumer.complete",
        return_value={"id_token": IdToken(iss="fake"), "access_token": "--"},
    )
    @mock.patch("django_pyoidc.client.Consumer.restore")
    @mock.patch("tests.e2e.test_app.callback.hook_validate_access_token")
    def test_callback_calling_hook_validate_access_token(
        self,
        mocked_user_access_token_hook,
        mocked_restore,
        mocked_complete,
        mocked_get_user_info,
        mocked_get_user,
        mocked_parse_authz,
    ):
        """
        Test that receiving a callback for a user that does not get validated by the developer-provided function 'get_user'
        does not get logged in
        """
        self.client.force_login(self.user)

        state = "test_id_12345"

        session = self.client.session
        session["oidc_sid"] = state
        session.save()

        # sso2 contains a definition with a hook for access token validation
        response = self.client.get(reverse("my_test_callback_sso2"))
        mocked_restore.assert_called_once_with(state)
        mocked_complete.assert_called_once_with(state=state, session_state=None)
        mocked_parse_authz.assert_called_once()
        mocked_get_user_info.assert_called_once_with(state=state)
        mocked_get_user.assert_called_once()

        self.assertRedirects(response, "/", fetch_redirect_response=False)
        self.assertEqual(OIDCSession.objects.all().count(), 0)
        mocked_user_access_token_hook.assert_called_once()


class BackchannelLogoutTestCase(OIDCTestCase):
    @classmethod
    def setUpTestData(cls):
        """
        To generate an other jwk key : 'jose jwk gen -i '{"alg":"HS256"}' -o oct.jwk'
        """
        cls.signing_key = jwk_from_dict(
            {
                "alg": "HS256",
                "k": "aMQ4QgzeE_XS91lxhixouomrhy_Tymz_xGC1dmwG8Vw",
                "key_ops": ["sign", "verify"],
                "kty": "oct",
            }
        )

    def test_invalid_backchannel_logout_wrong_method_request(self):
        """
        Test that performing a GET on a backchannel logout route results in an HTTP 405
        :return:
        """
        response = self.client.get(reverse("test_blogout"))
        self.assertEqual(response.status_code, 405)

    def test_invalid_backchannel_logout_empty_request(self):
        """
        Test that providing an empty body results in an HTTP 400
        :return:
        """
        response = self.client.post(
            reverse("test_blogout"), content_type="application/x-www-form-urlencoded"
        )
        self.assertEqual(response.status_code, 400)

    def test_invalid_encoding(self):
        """
        Test that using something else than 'application/x-www-form-urlencoded' for the content-type is rejected
        """
        response = self.client.post(
            reverse("test_blogout"),
            data="".encode("utf-8"),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 415)

    def test_invalid_backchannel_logout_no_sub_no_sid(self):
        """
        Test that providing an empty body results in an HTTP 400
        """

        payload = {}
        body = JWT().encode(payload, key=self.signing_key)
        request_body = "a" * 13 + body
        response = self.client.post(
            reverse("test_blogout"),
            data=request_body.encode("utf-8"),
            content_type="application/x-www-form-urlencoded",
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.content.decode("utf-8"),
            "Got invalid logout token : sub or sid is missing",
        )

    @mock.patch("django_pyoidc.views.SessionStore.delete")
    @mock.patch("django_pyoidc.client.Consumer.backchannel_logout")
    @mock.patch("django_pyoidc.client.Consumer.provider_config")
    def test_valid_backchannel_sub(
        self, mocked_provider_config, mocked_backchannel_logout, mocked_session_delete
    ):
        """
        Test that providing a valid SUB does kill the sessions
        """

        sub = "12333"
        state = "abcde"
        s = SessionStore()
        s["data"] = "data"
        s.create()
        cache_session_key = s.session_key

        OIDCSession.objects.create(
            sub=sub, cache_session_key=cache_session_key, state=state
        )

        payload = {"sub": sub}
        body = JWT().encode(payload, key=self.signing_key)
        request_body = "a" * 13 + body
        response = self.client.post(
            reverse("test_blogout"),
            data=request_body.encode("utf-8"),
            content_type="application/x-www-form-urlencoded",
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(OIDCSession.objects.all().count(), 0)
        mocked_backchannel_logout.assert_called_once()
        mocked_session_delete.assert_called_once_with(cache_session_key)

    @mock.patch("django_pyoidc.views.SessionStore.delete")
    @mock.patch("django_pyoidc.client.Consumer.backchannel_logout")
    @mock.patch("django_pyoidc.client.Consumer.provider_config")
    @mock.patch("django_pyoidc.client.Consumer.restore")
    def test_valid_backchannel_sid(
        self,
        mocked_restore,
        mocked_provider_config,
        mocked_backchannel_logout,
        mocked_session_delete,
    ):
        """
        Test that providing a valid SID does kill the session
        """

        sub = "12333"
        session_state = "sid:58"
        state = "abcde"
        s = SessionStore()
        s["data"] = "data"
        s.create()
        cache_session_key = s.session_key

        OIDCSession.objects.create(
            sub=sub,
            cache_session_key=cache_session_key,
            state=state,
            session_state=session_state,
        )

        mocked_backchannel_logout.return_value = session_state

        payload = {"sid": session_state}
        body = JWT().encode(payload, key=self.signing_key)
        request_body = "a" * 13 + body
        response = self.client.post(
            reverse("test_blogout"),
            data=request_body.encode("utf-8"),
            content_type="application/x-www-form-urlencoded",
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(OIDCSession.objects.all().count(), 0)
        mocked_backchannel_logout.assert_called_once()
        mocked_session_delete.assert_called_once_with(cache_session_key)
        mocked_restore.assert_called_once_with(session_state)

    @mock.patch("django_pyoidc.views.SessionStore.delete")
    @mock.patch("django_pyoidc.client.Consumer.backchannel_logout")
    @mock.patch("django_pyoidc.client.Consumer.provider_config")
    @mock.patch("django_pyoidc.client.Consumer.restore")
    def test_invalid_backchannel_sid(
        self,
        mocked_restore,
        mocked_provider_config,
        mocked_backchannel_logout,
        mocked_session_delete,
    ):
        """
        Test that providing a mismatching sid value results in an HTTP 400
        """

        sub = "12333"
        session_state = "sid:58"
        state = "abcde"
        s = SessionStore()
        s["data"] = "data"
        s.create()
        cache_session_key = s.session_key

        OIDCSession.objects.create(
            sub=sub,
            cache_session_key=cache_session_key,
            state=state,
            session_state=session_state,
        )

        mocked_backchannel_logout.return_value = "invalid_sid"

        payload = {"sid": session_state}
        body = JWT().encode(payload, key=self.signing_key)
        request_body = "a" * 13 + body
        response = self.client.post(
            reverse("test_blogout"),
            data=request_body.encode("utf-8"),
            content_type="application/x-www-form-urlencoded",
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(OIDCSession.objects.all().count(), 1)
        mocked_backchannel_logout.assert_called_once()
        mocked_session_delete.assert_not_called()
        mocked_restore.assert_called_once_with(session_state)

    @mock.patch("django_pyoidc.views.SessionStore.delete")
    @mock.patch("django_pyoidc.client.Consumer.backchannel_logout")
    @mock.patch("django_pyoidc.client.Consumer.provider_config")
    def test_valid_backchannel_sub_multiple_sessions(
        self, mocked_provider_config, mocked_backchannel_logout, mocked_session_delete
    ):
        """
        Test that providing a valid SUB does kill ALL the sessions
        """

        sub = "12333"
        s = SessionStore()
        s["data"] = "data1"
        s.create()
        cache_session_key_1 = s.session_key

        s = SessionStore()
        s["data"] = "data2"
        s.create()
        cache_session_key_2 = s.session_key

        OIDCSession.objects.create(
            sub=sub, cache_session_key=cache_session_key_1, state="1"
        )
        OIDCSession.objects.create(
            sub=sub, cache_session_key=cache_session_key_2, state="2"
        )

        payload = {"sub": sub}
        body = JWT().encode(payload, key=self.signing_key)
        request_body = "a" * 13 + body
        response = self.client.post(
            reverse("test_blogout"),
            data=request_body.encode("utf-8"),
            content_type="application/x-www-form-urlencoded",
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(OIDCSession.objects.all().count(), 0)
        mocked_backchannel_logout.assert_has_calls(
            [call(request_args=ANY), call(request_args=ANY)]
        )
        mocked_session_delete.assert_has_calls(
            [call(cache_session_key_1), call(cache_session_key_2)]
        )



import logging

from django.test import TestCase, override_settings

logger = logging.getLogger(__name__)


@override_settings(
    DJANGO_PYOIDC={
        "sso1": {
            "OIDC_CACHE_PROVIDER_METADATA": False,
            "client_id": "1",
            "client_secret": "--",
            "cache_django_backend": "default",
            "provider_class": "KeycloakProvider",
            "keycloak_base_uri": "http://sso1",
            "keycloak_realm": "realm1",
            "callback_uri_name": "my_test_callback",
            "post_login_uri_success": "/default/success",
            "login_uris_redirect_allowed_hosts": ["test.django-pyoidc.notatld"],
            "login_redirection_requires_https": True,
            "post_logout_redirect_uri": "/logoutdone",
            "post_login_uri_failure": "/logout_failure",
            "use_introspection_on_access_tokens": False,
        },
        "sso2": {
            "OIDC_CACHE_PROVIDER_METADATA": False,
            "client_id": "2",
            "client_secret": "--",
            "cache_django_backend": "default",
            "provider_discovery_uri": "http://sso2/uri",
            "login_uris_redirect_allowed_hosts": ["test2.django-pyoidc.notatld"],
            "oidc_callback_path": "callback-wtf/",
            "hook_validate_access_token": "tests.e2e.test_app.callback:hook_validate_access_token",
        },
        "sso3": {
            "oidc_cache_provider_metadata": True,
            "client_id": "3",
            "client_secret": "--",
            "cache_django_backend": "default",
            "provider_discovery_uri": "http://sso3/uri",
        },
        "sso4": {
            "OIDC_CACHE_PROVIDER_METADATA": True,
            "client_id": "4",
            "client_secret": "--",
            "cache_django_backend": "default",
            "provider_discovery_uri": "http://sso4/uri/.well-known/openid-configuration",
        },
        "drf": {
            "OIDC_CACHE_PROVIDER_METADATA": False,
            "client_id": "drf-api",
            "client_secret": "--",
            "provider_discovery_uri": "http://sso5/uri/.well-known/openid-configuration",
        },
        "sso10": {
            "OIDC_CACHE_PROVIDER_METADATA": False,
            "client_id": "10",
            "client_secret": "--",
            "callback_uri_name": "op10_namespace:sso10-callback",
            "cache_django_backend": "default",
            "provider_discovery_uri": "http://sso10/uri/.well-known/openid-configuration",
            "login_uris_redirect_allowed_hosts": ["test.django-pyoidc.notatld"],
        },
        "sso11": {
            "OIDC_CACHE_PROVIDER_METADATA": False,
            "client_id": "11",
            "client_secret": "--",
            "oidc_paths_prefix": "oidc11",
            "cache_django_backend": "default",
            "provider_discovery_uri": "http://sso11/uri/.well-known/openid-configuration",
            "login_uris_redirect_allowed_hosts": ["test.django-pyoidc.notatld"],
        },
        "sso12": {
            "OIDC_CACHE_PROVIDER_METADATA": False,
            "client_id": "12",
            "client_secret": "--",
            "oidc_paths_prefix": "oidc12-zz",
            "oidc_callback_path": "prefix12/oidc12-zz-callback",
            "cache_django_backend": "default",
            "provider_discovery_uri": "http://sso12/uri/.well-known/openid-configuration",
            "login_uris_redirect_allowed_hosts": ["test.django-pyoidc.notatld"],
        },
        "sso13": {
            "OIDC_CACHE_PROVIDER_METADATA": False,
            "client_id": "13",
            "client_secret": "--",
            "callback_uri_name": "op13_namespace:sso13-callback",
            "oidc_paths_prefix": "oidc13-ww",
            "cache_django_backend": "default",
            "provider_discovery_uri": "http://sso12/uri/.well-known/openid-configuration",
            "login_uris_redirect_allowed_hosts": ["test.django-pyoidc.notatld"],
        },
    }
)
class OIDCTestCase(TestCase):
    pass



import logging

from django.contrib import messages
from django.contrib.auth import get_user_model
from django.core.exceptions import PermissionDenied

from django_pyoidc.client import OIDCClient
from django_pyoidc.exceptions import ClaimNotFoundError
from django_pyoidc.utils import extract_claim_from_tokens

logger = logging.getLogger(__name__)


def login_callback(request, user):
    messages.success(request, f"login_callback for : {user.email}")


def logout_callback(request, logout_request_args):
    if request.user.is_anonymous:
        messages.success(request, "Logout Callback for anonymous user.")
    else:
        messages.success(request, f"Logout Callback for {request.user.email}.")

    logout_request_args["locale"] = "fr"
    logout_request_args["test"] = "zorg"
    return logout_request_args


def _debug_tokens(tokens={}):
    print("++++++++ TOKENS DEBUG ++++++++++++++++")
    if "info_token_claims" in tokens:
        print("info_token_claims")
        print(type(tokens["info_token_claims"]))  # dict
        print(tokens["info_token_claims"])

    if "access_token_jwt" in tokens:
        print("access_token_jwt")
        print(tokens["access_token_jwt"])
        print(type(tokens["access_token_jwt"]))  # <class 'str'>

    if "access_token_claims" in tokens:
        print("access_token_claims")
        print(
            type(tokens["access_token_claims"])
        )  # oic.extension.message.TokenIntrospectionResponse
        print(tokens["access_token_claims"])

    if "id_token_claims" in tokens:
        print("id_token_claims")
        print(type(tokens["id_token_claims"]))  # dict
        print(tokens["id_token_claims"])


def get_user_with_resource_access_check(client: OIDCClient, tokens={}):

    _debug_tokens(tokens)
    access_token_claims = (
        tokens["access_token_claims"] if "access_token_claims" in tokens else None
    )
    id_token_claims = tokens["id_token_claims"] if "id_token_claims" in tokens else None
    info_token_claims = (
        tokens["info_token_claims"] if "info_token_claims" in tokens else None
    )

    resource_access = (
        access_token_claims["resource_access"]
        if access_token_claims and "resource_access" in access_token_claims
        else None
    )

    # Perform resource access checks
    client_id = client.get_setting("client_id")
    # warning for user with no access Keycloak would not generate the resource_access claim
    # so we need to check absence of the claim also
    if (resource_access and client_id not in resource_access) or (
        resource_access is None
    ):
        logger.error("Failed resource access check in access_token")
        raise PermissionDenied("You do not have access to this application.")
    # then you could extend roles analysis if needed (for example this could need a local groups/roles mapping)
    # else
    # roles = resource_access[client_id]
    # if 'AccessApp1' in roles:
    #     do stuff

    username = ""

    if "preferred_username" in id_token_claims:
        username = id_token_claims["preferred_username"]
    elif "preferred_username" in info_token_claims:
        username = info_token_claims["preferred_username"]
    else:
        username = info_token_claims["email"]

    User = get_user_model()
    user, created = User.objects.get_or_create(
        email=info_token_claims["email"],
        username=username,
    )
    user.is_superuser = (
        "groups" in info_token_claims and "admins" in info_token_claims["groups"]
    )
    user.backend = "django.contrib.auth.backends.ModelBackend"
    user.save()

    return user


def extract_username_from_tokens(tokens: dict):
    try:
        username = extract_claim_from_tokens("preferred_username", tokens)
    except ClaimNotFoundError:
        try:
            username = extract_claim_from_tokens("email", tokens)
        except ClaimNotFoundError:
            raise ClaimNotFoundError(
                "We found nothing to extract a username in current OIDC tokens."
            )
    return username


def get_user_with_minimal_audiences_check(client: OIDCClient, tokens={}):
    """Checking audiences on the access token. Works with access_token_claims extracted only."""
    _debug_tokens(tokens)
    access_token_claims = (
        tokens["access_token_claims"]
        if "access_token_claims" in tokens and tokens["access_token_claims"]
        else None
    )
    # id_token_claims = tokens["id_token_claims"] if "id_token_claims" in tokens else None
    info_token_claims = (
        tokens["info_token_claims"] if "info_token_claims" in tokens else None
    )

    # Perform a minimal audience check
    # Note: here not checking if client_id is in 'aud' because that's broken in Keycloak
    client_id = client.get_setting("client_id")
    if access_token_claims:
        if "azp" not in access_token_claims:
            logger.error("Missing azp claim access_token")
            raise PermissionDenied("You do not have access to this application.")
        elif not access_token_claims["azp"] == client_id:
            logger.error("Failed audience (azp claim) minimal check in access_token")
            raise PermissionDenied("You do not have access to this application.")
    else:
        logger.error(
            "Missing access_token claims. This function need an extracted access token to work, please activate 'use_introspection_on_access_tokens' or 'hook_validate_access_token'."
        )
        raise ClaimNotFoundError(
            "get_user_with_minimal_audiences_check works only with claims extracted from the access token."
        )

    username = extract_username_from_tokens(tokens)

    User = get_user_model()
    user, created = User.objects.get_or_create(
        email=info_token_claims["email"],
        username=username,
    )
    user.is_superuser = (
        "groups" in info_token_claims and "admins" in info_token_claims["groups"]
    )
    user.backend = "django.contrib.auth.backends.ModelBackend"
    user.save()

    return user


def get_user_with_audiences_check(client: OIDCClient, tokens={}):

    _debug_tokens(tokens)
    access_token_claims = (
        tokens["access_token_claims"] if "access_token_claims" in tokens else None
    )
    # id_token_claims = tokens["id_token_claims"] if "id_token_claims" in tokens else None
    info_token_claims = (
        tokens["info_token_claims"] if "info_token_claims" in tokens else None
    )

    # Note: Keycloak broke the audience checks for access_tokens, an access token generated by "X" request
    # will not contain "X" in the 'aud' key, only extra audiences.
    # so with keycloak you can only perform audiences checks when receiving access tokens generated
    # by other means (like by a front)
    audiences = []
    if "aud" in access_token_claims:
        if type(access_token_claims["aud"]) is str:
            audiences = [
                access_token_claims["aud"],
            ]
        elif type(access_token_claims["aud"]) is bytes:
            audiences = [
                access_token_claims["aud"].decode("ascii"),
            ]
        elif type(access_token_claims["aud"]) is list:
            audiences = access_token_claims["aud"]
        else:
            raise RuntimeError("Unknown type for audience claim")

    # Perform audience check
    if audiences and client.get_setting("client_id") not in audiences:
        logger.error("Failed audience check in access_token")
        raise PermissionDenied("You do not have access to this application.")

    username = extract_username_from_tokens(tokens)

    User = get_user_model()
    user, created = User.objects.get_or_create(
        email=info_token_claims["email"],
        username=username,
    )
    user.is_superuser = (
        "groups" in info_token_claims and "admins" in info_token_claims["groups"]
    )
    user.backend = "django.contrib.auth.backends.ModelBackend"
    user.save()

    return user


def hook_validate_access_token(access_token_jwt: str, client: OIDCClient):
    """Here we should handle the jwt string and convert it to a claim dictionnary.

    But we should also ensure validity of the jwt (like signatures, expiration, etc.).
    """
    return {}



import logging
from importlib import import_module
from typing import Any, Dict, Optional, TypeVar, Union

# import oic
from django.conf import settings
from django.contrib import auth, messages
from django.contrib.auth.models import AbstractUser
from django.core.exceptions import PermissionDenied, SuspiciousOperation
from django.http import HttpRequest, HttpResponse
from django.shortcuts import redirect, resolve_url
from django.utils.decorators import method_decorator
from django.utils.http import url_has_allowed_host_and_scheme
from django.views import View
from django.views.decorators.csrf import csrf_exempt
from jwt import JWT
from jwt.exceptions import JWTDecodeError
from oic.utils.http_util import BadRequest

from django_pyoidc.client import OIDCClient
from django_pyoidc.engine import OIDCEngine
from django_pyoidc.exceptions import InvalidSIDException
from django_pyoidc.models import OIDCSession
from django_pyoidc.settings import OIDCSettings, OIDCSettingsFactory, OidcSettingValue
from django_pyoidc.utils import import_object

SessionStore = import_module(settings.SESSION_ENGINE).SessionStore

logger = logging.getLogger(__name__)

T = TypeVar("T")


class OIDCMixin:
    op_name: str = ""
    opsettings: OIDCSettings


class OIDCView(View, OIDCMixin):
    def __init__(self, **kwargs: Any) -> None:
        for key, value in kwargs.items():
            setattr(self, key, value)
            if key == "op_name":
                self.opsettings = OIDCSettingsFactory.get(self.op_name)
                self.allowed_hosts = self.get_setting(
                    "login_uris_redirect_allowed_hosts"
                )

    def setup(self, request: HttpRequest, *args: Any, **kwargs: Any) -> None:
        super().setup(request, *args, **kwargs)
        if self.op_name is None:
            raise Exception(
                "Please set 'op_name' when initializing with 'as_view()'\nFor example : OIDCView.as_view(op_name='example')"
            )  # FIXME

    def get_setting(
        self, name: str, default: Optional[T] = None
    ) -> Optional[Union[T, OidcSettingValue]]:
        return self.opsettings.get(name, default)

    def call_function(self, setting_func_name: str, *args: Any, **kwargs: Any) -> Any:
        function_path: Optional[str] = self.opsettings.get(setting_func_name)  # type: ignore[assignment] # we can assume that the configuration is right
        if function_path is not None:
            func = import_object(function_path, "")
            return func(*args, **kwargs)

    def call_user_login_callback_function(
        self, request: HttpRequest, user: AbstractUser
    ) -> Any:
        logger.debug("OIDC, Calling user hook on login")
        self.call_function("hook_user_login", request, user)

    def call_logout_function(
        self, user_request: HttpRequest, logout_request_args: Dict[str, Any]
    ) -> Any:
        """Function called right before local session removal and before final redirection to the SSO server.

        Parameters:
            user_request : current user request on Django
            logout_request_args : current arguments added to the SSO server logout link. The final list of arguments
              is made by pyoidc and will add the id_token_hint element

        Returns:
            dict: extra query string arguments to add to the SSO disconnection url
        """
        return self.call_function("hook_user_logout", user_request, logout_request_args)

    def get_next_url(
        self, request: HttpRequest, redirect_field_name: str
    ) -> Optional[str]:
        """
        Adapted from https://github.com/mozilla/mozilla-django-oidc/blob/71e4af8283a10aa51234de705d34cd298e927f97/mozilla_django_oidc/views.py#L132
        """
        next_url = request.GET.get(redirect_field_name)
        if next_url:
            is_safe = url_has_allowed_host_and_scheme(
                next_url,
                allowed_hosts=self.allowed_hosts,  # type: ignore[arg-type] # let's just assume that this settings is correctly set
                require_https=self.get_setting(  # type: ignore[arg-type] # We can reasonably assume that this setting is a bool
                    "login_redirection_requires_https", True
                ),
            )
            if is_safe:
                return request.build_absolute_uri(next_url)
        return None


class OIDCLoginView(OIDCView):
    """
    When receiving a GET request, this views redirects the user to the SSO identified by `op_name`.
    This view is named ``<op_name>-login`` if you used ``get_urlpatterns``.

    This view supports the *http query parameter* ``next`` (ie ``?next=http://...``) to specify which url the user should be redirected to on success.

    First, an OIDC redirection is made to the sso, with a callback (redirection) set to a local url defined by the setting:

    * :ref:`oidc_callback_path` local path to be redirected after authentication on the sso, to finalize the local auth.

    After this somewhat internal redirection where the local auth is validated and the session created, a final redirection
    will be made.
    The final redirection behaviour is configured with the following settings :

    * :ref:`login_redirection_requires_https` controls if non https URIs are accepted.
    * :ref:`login_uris_redirect_allowed_hosts` controls which hosts the user can be redirected to.
    * :ref:`post_login_uri_success` defines the redirection URI when no 'next' redirect uri were provided in the HTTP request.
    """

    http_method_names = ["get"]

    def get(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:

        sid = request.session.get("oidc_sid")
        if sid:
            try:
                client = OIDCClient(self.op_name, session_id=sid)
            except InvalidSIDException:
                # maybe a failed attempt trace in the session.
                # we ignore the session sid and get back on the first steps.
                client = OIDCClient(self.op_name)
        else:
            client = OIDCClient(self.op_name)

        callback_path = str(self.get_setting("oidc_callback_path"))
        if not callback_path.startswith("/"):
            # issue in pyoidc when base_url ends with "/" and path does not start with "/" there's an extra "/" added (doubling)
            # to prevent that we finally enforce the "/" prefix on the path
            callback_path = f"/{callback_path}"
        client.consumer.consumer_config["authz_page"] = callback_path
        next_redirect_uri = self.get_next_url(request, "next")

        if not next_redirect_uri:
            next_redirect_uri = str(
                self.get_setting(
                    "post_login_uri_success", request.build_absolute_uri("/")
                )
            )

        request.session["oidc_login_next"] = next_redirect_uri

        sid, location = client.consumer.begin(  # type: ignore[no-untyped-call] # oic package is untyped
            scope=["openid"],
            response_type="code",
            use_nonce=True,
            path=self.request.build_absolute_uri("/"),
        )
        request.session["oidc_sid"] = sid
        return redirect(location)


class OIDCLogoutView(OIDCView):
    """
    This view logs out the user, killing it's session on this service and notifying the identity provider that it has logged-out.
    It is named ``<op_name>-logout`` if you used ``get_urlpatterns``.

    It supports both ``GET`` and ``POST`` http methods.

    The response is a redirection to the SSO logout endpoint, if a provider configuration :ref:`post_logout_redirect_uri` exists it as used as
    post logout redirection argument on the SSO redirection link.

    """

    http_method_names = ["get", "post"]

    def post_logout_url(self, request: HttpRequest) -> str:
        """Return the post logout url defined in settings."""
        return str(
            self.get_setting(
                "post_logout_redirect_uri", request.build_absolute_uri("/")
            )
        )

    def get(self, request: HttpRequest) -> HttpResponse:
        return self.post(request)

    def post(self, request: HttpRequest) -> HttpResponse:
        """Log out the user."""
        url = self.post_logout_url(request)
        # If this url is not already an absolute url
        # we make it absolute using the current domain
        if not url[:7] in ["http://", "https:/"]:
            post_logout_url = request.build_absolute_uri(url)
        else:
            post_logout_url = url

        if not request.user.is_authenticated:
            redirect(post_logout_url)

        client = None
        sid = request.session.get("oidc_sid")

        redirect_arg_name: str = self.get_setting(
            "LOGOUT_QUERY_STRING_REDIRECT_PARAMETER",
            "post_logout_redirect_uri",
        )  # type: ignore[assignment] # we can assume that the configuration is right
        request_args = {
            redirect_arg_name: post_logout_url,
            "client_id": self.get_setting("client_id"),
        }

        # Allow some more parameters for some actors
        extra_logout_args: Dict[str, Any] = self.get_setting(  # type: ignore[assignment] # we can assume that the configuration is right
            "oidc_logout_query_string_extra_parameters_dict",
            {},
        )
        request_args.update(extra_logout_args)
        if sid:
            try:
                client = OIDCClient(self.op_name, session_id=sid)
            except (
                Exception
            ) as e:  # FIXME : Finer exception handling (KeyError,ParseError,CommunicationError)
                logger.error("OIDC Logout call error when loading OIDC state: ")
                logger.exception(e)

        # Hook user logout function
        extra_args = self.call_logout_function(request, request_args)
        if extra_args:
            extra_args.update(request_args)
        else:
            extra_args = request_args

        # Django disconnection
        auth.logout(request)

        if sid:
            OIDCSession.objects.filter(session_state=sid).delete()

        # Build SSO disconnect link
        if client:
            logout_request = client.consumer.message_factory.get_request_type(
                "endsession_endpoint"
            )
            url, body, http_args, _ = client.consumer.request_info(
                logout_request,
                method="GET",
                request_args=None,
                extra_args=extra_args,
                scope="",
                state=sid,
                prop="id_token_hint",
            )
            logger.debug("SSO logout: endsession redirect url: %s", url)
            return redirect(url)
        else:
            logger.error(
                f"Cannot build the SSO disconnection link (maybe the cache was flushed ?), still redirecting directly to {post_logout_url}"
            )
            return redirect(post_logout_url)


@method_decorator(csrf_exempt, name="dispatch")
class OIDCBackChannelLogoutView(OIDCView):
    """
    This view only accept POST requests. This is where your identity provider notifies the library that we should kill a user
    session. Usually, you should not redirect a user manually to this view.

    It is named ``<op_name>-backchannel-logout`` if you used ``get_urlpatterns``.
    """

    http_method_names = ["post"]

    def logout_sessions_by_sid(self, client: OIDCClient, sid: str, body: str) -> None:
        validated_sid = client.consumer.backchannel_logout(
            request_args={"logout_token": body}
        )
        if validated_sid != sid:
            raise InvalidSIDException(f"Got {validated_sid}, expected {sid}")
        sessions = OIDCSession.objects.filter(session_state=validated_sid)
        for session in sessions:
            self._logout_session(session)

    def logout_sessions_by_sub(self, client: OIDCClient, sub: str, body: str) -> None:
        sessions = OIDCSession.objects.filter(sub=sub)
        for session in sessions:
            client.consumer.backchannel_logout(request_args={"logout_token": body})
            self._logout_session(session)

    def _logout_session(self, session: OIDCSession) -> None:
        s = SessionStore()
        s.delete(session.cache_session_key)
        session.delete()
        logger.info(f"Backchannel logout request received and validated for {session}")

    def post(self, request: HttpRequest) -> HttpResponse:
        if request.content_type != "application/x-www-form-urlencoded":
            return HttpResponse("", status=415)
        result = HttpResponse("")
        try:
            body = request.body.decode("utf-8")[13:]
            decoded = JWT().decode(body, do_verify=False)  # type: ignore[no-untyped-call] # jwt.JWT is not typed yet

            sid = decoded.get("sid")
            sub = decoded.get("sub")
            if sub:
                # Authorization server wants to kill all sessions
                client = OIDCClient(self.op_name)
                self.logout_sessions_by_sub(client, sub, body)
            elif sid:
                client = OIDCClient(self.op_name, session_id=sid)
                try:
                    self.logout_sessions_by_sid(client, sid, body)
                except InvalidSIDException as e:
                    logger.debug(
                        f"Got invalid sid from request : expected {sid}. Error : \n{e}"
                    )
                    result.status_code = 400
            else:
                result.status_code = 400
                result.content = "Got invalid logout token : sub or sid is missing"
                logger.debug("Got invalid logout token : sub or sid is missing")
        except JWTDecodeError:
            result.status_code = 400
        except UnicodeDecodeError as e:
            raise SuspiciousOperation(e)
        result.headers["Cache-Control"] = "no-store"
        return result


class OIDCCallbackView(OIDCView):
    """
    This view only accepts GET request. This is where the identity provider redirects the user in the *Authorization Code Flow*.
    Usually, you should not redirect a user manually to this view.

    It is named ``<op_name>-callback`` if you used ``get_urlpatterns``.
    """

    http_method_names = ["get"]

    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)
        self.engine = OIDCEngine(self.opsettings)

    def success_url(self, request: HttpRequest) -> str:
        # Pull the next url from the session or settings --we don't need to
        # sanitize here because it should already have been sanitized.
        next_url = self.request.session.get("oidc_login_next", None)
        return next_url or resolve_url(
            self.get_setting("post_login_uri_success", request.build_absolute_uri("/"))  # type: ignore[arg-type] # we can assume that this setting is correctly configured
        )

    def login_failure(self, request: HttpRequest) -> HttpResponse:
        return redirect(
            str(
                self.get_setting(
                    "post_login_uri_failure", request.build_absolute_uri("/")
                )
            )
        )

    def get(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
        try:
            if "oidc_sid" in request.session:
                self.client = OIDCClient(
                    self.op_name, session_id=request.session["oidc_sid"]
                )

                parsing_result = self.client.consumer.parse_authz(
                    query=request.GET.urlencode()
                )
                if isinstance(parsing_result, BadRequest):
                    logger.error(
                        "OIDC login process failure; cannot parse OIDC response"
                    )
                    return self.login_failure(request)

                aresp, atr, idt = parsing_result

                if aresp is None:
                    logger.error("OIDC login process failure; empty OIDC response")
                    return self.login_failure(request)

                if aresp["state"] == request.session["oidc_sid"]:
                    state = aresp["state"]
                    session_state = aresp.get("session_state")  # type: ignore[no-untyped-call] # oic is untyped yet

                    # pyoidc will make the next steps in OIDC login protocol
                    try:
                        tokens = self.client.consumer.complete(
                            state=state, session_state=session_state
                        )
                    except Exception as e:
                        logger.exception(e)
                        logger.error(
                            "OIDC login process failure; cannot end login protocol."
                        )
                        return self.login_failure(request)

                    # Collect data from userinfo endpoint
                    try:
                        userinfo = self.client.consumer.get_user_info(state=state)  # type: ignore[no-untyped-call] # oic is untyped yet
                    except Exception as e:
                        logger.exception(e)
                        logger.error(
                            "OIDC login process failure; Cannot retrieve userinfo."
                        )
                        return self.login_failure(request)

                    # TODO: add a setting to allow/disallow session storage of the tokens
                    access_token_jwt = (
                        tokens["access_token"] if "access_token" in tokens else None
                    )

                    # this will call token instrospection or user defined validator
                    # or return None
                    access_token_claims = self.engine.introspect_access_token(
                        access_token_jwt, self.client
                    )

                    id_token_claims = (
                        tokens["id_token"].to_dict() if "id_token" in tokens else None
                    )
                    # id_token_jwt = (
                    #     tokens["id_token_jwt"] if "id_token_jwt" in tokens else None
                    # )
                    userinfo_claims = userinfo.to_dict()
                    tokens = {
                        "info_token_claims": userinfo_claims,
                        "access_token_jwt": access_token_jwt,
                        "access_token_claims": access_token_claims,
                        "id_token_claims": id_token_claims,
                    }
                    # simplify check code, if any dict is None remove the entry
                    filtered_tokens = {k: v for k, v in tokens.items() if v is not None}
                    # Call user hook
                    user = self.engine.call_get_user_function(
                        tokens=filtered_tokens,
                        client=self.client,
                    )

                    if not user or not user.is_authenticated:
                        logger.error(
                            "OIDC login process failure. Cannot set active authenticated user."
                        )
                        return self.login_failure(request)
                    else:
                        auth.login(request, user)
                        OIDCSession.objects.create(
                            state=state,
                            sub=userinfo["sub"],
                            cache_session_key=request.session.session_key,  # type: ignore[misc] # we call auth.login right before, so session_key is set to a value
                            session_state=session_state,
                        )
                        self.call_user_login_callback_function(request, user)
                        redir = self.success_url(request)
                        return redirect(redir)
                else:
                    logger.warning(
                        "OIDC login process failure. OIDC state does not match session sid."
                    )
                    raise SuspiciousOperation(
                        "Login process: OIDC state does not match session sid."
                    )
            else:
                logger.warning(
                    "OIDC login process failure. No OIDC sid state in user session for a request on the OIDC callback."
                )
                return self.login_failure(request)
        except PermissionDenied as exc:
            logger.exception(exc)
            messages.error(request, "Permission Denied.")
            return self.login_failure(request)



from typing import Any, List

from django.urls import URLPattern, path

from django_pyoidc.settings import OIDCSettings, OIDCSettingsFactory


class OIDCHelper:
    """
    This is a utility class providing a wrapper around the provider, the settings and the views.
    """

    def __init__(self, *args: Any, op_name: str, **kwargs: Any):
        """
        Parameters:
            op_name (str): the name of the sso provider that you are using.
              This should exists as a key in the DJANGO_PYOIDC settings section.
        """
        self.op_name = op_name
        self.opsettings: OIDCSettings = OIDCSettingsFactory.get(self.op_name)
        self.provider = self.opsettings.provider

    @property
    def login_uri_name(self) -> str:
        """
        The login viewname (use in :func:`django:django.urls.reverse` django directive for example) of this configuration
        """
        return f"{self.op_name}-login"

    @property
    def logout_uri_name(self) -> str:
        """
        The logout viewname (use in :func:`django:django.urls.reverse` django directive for example) of this configuration
        """
        return f"{self.op_name}-logout"

    @property
    def callback_uri_name(self) -> str:
        """
        The callback viewname (use in :func:`django:django.urls.reverse` django directive for example) of this configuration
        """
        return f"{self.op_name}-callback"

    @property
    def backchannel_logout_uri_name(self) -> str:
        """
        The backchannel logout viewname (use in :func:`django:django.urls.reverse` django directive for example) of this configuration
        """
        return f"{self.op_name}-backchannel-logout"

    def get_urlpatterns(self) -> List[URLPattern]:
        """
        Returns:
            A list of urllpatterns to be included using :func:`django:django.urls.include` in your url configuration
        """
        from django_pyoidc.views import (
            OIDCBackChannelLogoutView,
            OIDCCallbackView,
            OIDCLoginView,
            OIDCLogoutView,
        )

        oidc_paths_prefix = self.opsettings.get("oidc_paths_prefix")
        result = [
            path(
                f"{oidc_paths_prefix}-callback",
                OIDCCallbackView.as_view(op_name=self.op_name),
                name=self.callback_uri_name,
            ),
            path(
                f"{oidc_paths_prefix}-login",
                OIDCLoginView.as_view(op_name=self.op_name),
                name=self.login_uri_name,
            ),
            path(
                f"{oidc_paths_prefix}-logout",
                OIDCLogoutView.as_view(op_name=self.op_name),
                name=self.logout_uri_name,
            ),
            path(
                f"{oidc_paths_prefix}-backchannel-logout",
                OIDCBackChannelLogoutView.as_view(op_name=self.op_name),
                name=self.backchannel_logout_uri_name,
            ),
        ]
        return result



from unittest import mock
from unittest.mock import call

from oic.oauth2 import ASConfigurationResponse

from django_pyoidc.client import OIDCClient
from django_pyoidc.settings import OIDCSettingsFactory
from django_pyoidc.utils import OIDCCacheBackendForDjango
from tests.utils import OIDCTestCase


class CacheTestCase(OIDCTestCase):
    @mock.patch("django_pyoidc.client.Consumer.provider_config")
    def test_providers_info_not_cached(self, mocked_provider_config):
        """
        Test that multiple Clients creation without cache means several provider_info calls
        """
        # OIDCClient creation generates one call to Consumer.provider_config
        sso1 = OIDCClient(op_name="sso1")
        mocked_provider_config.assert_has_calls([call("http://sso1/realms/realm1")])
        assert 1 == mocked_provider_config.call_count

        # restoring a user session does not create a new call
        sso1.consumer._backup(sid="1234")
        OIDCClient(op_name="sso1", session_id="1234")
        mocked_provider_config.assert_has_calls([call("http://sso1/realms/realm1")])
        assert 1 == mocked_provider_config.call_count

        # but a new empty Client would add a new metadata call
        OIDCClient(op_name="sso1")
        mocked_provider_config.assert_has_calls(
            [call("http://sso1/realms/realm1"), call("http://sso1/realms/realm1")]
        )
        assert 2 == mocked_provider_config.call_count

    @mock.patch(
        "django_pyoidc.client.Consumer.provider_config",
        return_value=ASConfigurationResponse(),
    )
    def test_providers_info_cached(self, mocked_provider_config):
        """
        Test that multiple Clients creation with cache means only one provider_info call.
        """
        # empty the caches
        settings1 = OIDCSettingsFactory.get("sso3")
        settings2 = OIDCSettingsFactory.get("sso4")
        cache1 = OIDCCacheBackendForDjango(settings1)
        cache1.clear()
        cache2 = OIDCCacheBackendForDjango(settings2)
        cache2.clear()

        # OIDCClient creation generates one call to Consumer.provider_config
        sso1 = OIDCClient(op_name="sso3")
        mocked_provider_config.assert_has_calls([call("http://sso3/uri")])
        assert 1 == mocked_provider_config.call_count

        # restoring a user session does not create a new call
        sso1.consumer._backup(sid="1234")
        OIDCClient(op_name="sso3", session_id="1234")
        mocked_provider_config.assert_has_calls([call("http://sso3/uri")])
        assert 1 == mocked_provider_config.call_count

        # creating a new OIDCClient should activate shared cache
        # and prevent a new call
        OIDCClient(op_name="sso3")
        mocked_provider_config.assert_has_calls([call("http://sso3/uri")])
        assert 1 == mocked_provider_config.call_count

        # BUT adding a new Client with a different op_name will add a call,
        # as it is not the same cache key
        OIDCClient(op_name="sso4")

        mocked_provider_config.assert_has_calls(
            [call("http://sso3/uri"), call("http://sso4/uri")]
        )
        assert 2 == mocked_provider_config.call_count



from django.contrib.auth.models import User
from django.urls import include, path
from rest_framework import routers, serializers, viewsets
from rest_framework.permissions import AllowAny

from django_pyoidc.helper import OIDCHelper
from django_pyoidc.views import (
    OIDCBackChannelLogoutView,
    OIDCCallbackView,
    OIDCLoginView,
    OIDCLogoutView,
)
from tests.e2e.test_app.models import Public
from tests.e2e.test_app.views import (
    OIDCTestFailureView,
    OIDCTestLogoutView,
    OIDCTestSuccessView,
)


# DRF tests ----
# Serializers define the API representation.
class UserSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = User
        fields = ["url", "username", "email", "is_staff"]


class PublicSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = Public
        fields = ["data"]


# ViewSets define the view behavior.
class UserViewSet(viewsets.ModelViewSet):
    queryset = User.objects.all()
    serializer_class = UserSerializer


# ViewSets define the view behavior.
class PublicViewSet(viewsets.ModelViewSet):
    queryset = Public.objects.all()
    permission_classes = [AllowAny]
    serializer_class = PublicSerializer


# Routers provide an easy way of automatically determining the URL conf.
apirouter = routers.DefaultRouter()
apirouter.register(r"users", UserViewSet)
apirouter.register(r"publics", PublicViewSet)

helper_lemon1 = OIDCHelper(op_name="lemon1")
urlpatterns = [
    path("login-1/", OIDCLoginView.as_view(op_name="sso1"), name="e2e_test_login_1"),
    path("login-2/", OIDCLoginView.as_view(op_name="sso2"), name="e2e_test_login_2"),
    path("login-3/", OIDCLoginView.as_view(op_name="sso3"), name="e2e_test_login_3"),
    path("login-4/", OIDCLoginView.as_view(op_name="sso4"), name="e2e_test_login_4"),
    path("login-5/", OIDCLoginView.as_view(op_name="sso5"), name="e2e_test_login_5"),
    path(
        "callback-1/",
        OIDCCallbackView.as_view(op_name="sso1"),
        name="e2e_test_callback_1",
    ),
    path(
        "callback-2/",
        OIDCCallbackView.as_view(op_name="sso2"),
        name="e2e_test_callback_2",
    ),
    path(
        "callback-3/",
        OIDCCallbackView.as_view(op_name="sso3"),
        name="e2e_test_callback_3",
    ),
    path(
        "callback-4/",
        OIDCCallbackView.as_view(op_name="sso4"),
        name="e2e_test_callback_4",
    ),
    path(
        "callback-5/",
        OIDCCallbackView.as_view(op_name="sso5"),
        name="e2e_test_callback_5",
    ),
    path(
        "prefix_lemon1/",
        include(
            (helper_lemon1.get_urlpatterns(), "django_pyoidc"),
            namespace="lemon1_namespace",
        ),
    ),
    path(
        "test-ll-success-1/",
        OIDCTestSuccessView.as_view(op_name="lemon1"),
        name="e2e_test_ll_success_1",
    ),
    path(
        "test-ll-failure-1/",
        OIDCTestFailureView.as_view(op_name="lemon1"),
        name="e2e_test_ll_failure_1",
    ),
    path(
        "test-ll-logout-done-1/",
        OIDCTestLogoutView.as_view(op_name="lemon1"),
        name="e2e_test_ll_logout_done_1",
    ),
    path(
        "logout-1/",
        OIDCLogoutView.as_view(op_name="sso1"),
        name="e2e_test_logout_1",
    ),
    path(
        "logout-2/",
        OIDCLogoutView.as_view(op_name="sso2"),
        name="e2e_test_logout_2",
    ),
    path(
        "logout-3/",
        OIDCLogoutView.as_view(op_name="sso3"),
        name="e2e_test_logout_3",
    ),
    path(
        "logout-4/",
        OIDCLogoutView.as_view(op_name="sso4"),
        name="e2e_test_logout_4",
    ),
    path(
        "logout-5/",
        OIDCLogoutView.as_view(op_name="sso5"),
        name="e2e_test_logout_5",
    ),
    path(
        "back_channel_logout-1/",
        OIDCBackChannelLogoutView.as_view(op_name="sso1"),
        name="e2e_test_blogout_1",
    ),
    path(
        "back_channel_logout-2/",
        OIDCBackChannelLogoutView.as_view(op_name="sso2"),
        name="e2e_test_blogout_2",
    ),
    path(
        "back_channel_logout-3/",
        OIDCBackChannelLogoutView.as_view(op_name="sso3"),
        name="e2e_test_blogout_3",
    ),
    path(
        "back_channel_logout-4/",
        OIDCBackChannelLogoutView.as_view(op_name="sso4"),
        name="e2e_test_blogout_4",
    ),
    path(
        "back_channel_logout-5/",
        OIDCBackChannelLogoutView.as_view(op_name="sso5"),
        name="e2e_test_blogout_5",
    ),
    path(
        "test-success-1/",
        OIDCTestSuccessView.as_view(op_name="sso1"),
        name="e2e_test_success_1",
    ),
    path(
        "test-success-2/",
        OIDCTestSuccessView.as_view(op_name="sso2"),
        name="e2e_test_success_2",
    ),
    path(
        "test-success-3/",
        OIDCTestSuccessView.as_view(op_name="sso3"),
        name="e2e_test_success_3",
    ),
    path(
        "test-success-4/",
        OIDCTestSuccessView.as_view(op_name="sso4"),
        name="e2e_test_success_4",
    ),
    path(
        "test-success-5/",
        OIDCTestSuccessView.as_view(op_name="sso5"),
        name="e2e_test_success_5",
    ),
    path(
        "test-logout-done-1/",
        OIDCTestLogoutView.as_view(op_name="sso1"),
        name="e2e_test_logout_done_1",
    ),
    path(
        "test-logout-done-2/",
        OIDCTestLogoutView.as_view(op_name="sso2"),
        name="e2e_test_logout_done_2",
    ),
    path(
        "test-logout-done-3/",
        OIDCTestLogoutView.as_view(op_name="sso3"),
        name="e2e_test_logout_done_3",
    ),
    path(
        "test-logout-done-4/",
        OIDCTestLogoutView.as_view(op_name="sso4"),
        name="e2e_test_logout_done_4",
    ),
    path(
        "test-logout-done-5/",
        OIDCTestLogoutView.as_view(op_name="sso5"),
        name="e2e_test_logout_done_5",
    ),
    path(
        "test-failure-1/",
        OIDCTestFailureView.as_view(op_name="sso1"),
        name="e2e_test_failure_1",
    ),
    path(
        "test-failure-2/",
        OIDCTestFailureView.as_view(op_name="sso2"),
        name="e2e_test_failure_2",
    ),
    path(
        "test-failure-3/",
        OIDCTestFailureView.as_view(op_name="sso3"),
        name="e2e_test_failure_3",
    ),
    path(
        "test-failure-4/",
        OIDCTestFailureView.as_view(op_name="sso4"),
        name="e2e_test_failure_4",
    ),
    path(
        "test-failure-5/",
        OIDCTestFailureView.as_view(op_name="sso5"),
        name="e2e_test_failure_5",
    ),
    path("api/", include(apirouter.urls)),
]



class InvalidSIDException(Exception):
    pass


class TokenError(Exception):
    pass


class ExpiredToken(Exception):
    pass


class ClaimNotFoundError(Exception):
    pass


class InvalidOIDCConfigurationException(Exception):
    pass



from django_pyoidc.providers.keycloak_18 import Keycloak18Provider


class KeycloakProvider(Keycloak18Provider):
    """
    Provide django settings/urlconf based on keycloak behaviour (latest version).

    For older Keycloak versions please check the other Keycloak_* providers.
    """

    pass



from importlib import import_module

from django.conf import settings
from django.contrib import admin

from django_pyoidc.models import OIDCSession

SessionStore = import_module(settings.SESSION_ENGINE).SessionStore


class OIDCSessionAdmin(admin.ModelAdmin):  # type: ignore[type-arg] # https://github.com/typeddjango/django-stubs/issues/507
    readonly_fields = (
        "state",
        "session_state",
        "has_session_management",
        "sub",
        "cache_session_key",
        "session_is_active",
        "created_at",
    )
    list_display = [
        "id",
        "has_session_management",
        "session_is_active",
        "sub",
        "created_at",
    ]

    @admin.display(boolean=True)
    def has_session_management(self, obj: OIDCSession) -> bool:
        return obj.session_state is not None

    @admin.display(boolean=True)
    def session_is_active(self, obj: OIDCSession) -> bool:
        s = SessionStore()
        return obj.cache_session_key is not None and s.exists(obj.cache_session_key)



"""
Base Keycloak Provider class.
"""

from typing import Any, Optional
from urllib.parse import urlparse

from typing_extensions import override

from django_pyoidc.providers.provider import Provider, ProviderConfig


class Keycloak10Provider(Provider):
    """
    Provide Django settings/urlconf based on keycloak behaviour (v10 to v18)
    """

    def __init__(
        self,
        keycloak_base_uri: Optional[str] = None,
        keycloak_realm: Optional[str] = None,
        *args: Any,
        op_name: str,
        **kwargs: Any,
    ):
        if keycloak_base_uri is None or keycloak_realm is None:
            # Usage of this provider SHOULD be providing keycloak_base_uri and keycloak_realm so we generate provider_discovery_uri
            # but the contrary works (you provide provider_discovery_uri and we extract base_uri and realm).
            if (
                "provider_discovery_uri" not in kwargs
                or not kwargs["provider_discovery_uri"]
            ):
                raise TypeError(
                    "Keycloak10Provider requires keycloak_base_uri and keycloak_realm or provider_discovery_uri."
                )
            url = urlparse(kwargs["provider_discovery_uri"])
            base_path = url.path
            if "/realms/" in base_path:
                parts = base_path.split("/realms/")
                base_path = parts[0]
                keycloak_realm = parts[1]
                extra_string = ".well-known/openid-configuration"
                if keycloak_realm is not None and keycloak_realm.endswith(extra_string):
                    keycloak_realm = keycloak_realm[: -len(extra_string)]
                extra_string = ".well-known/openid-configuration/"
                if keycloak_realm is not None and keycloak_realm.endswith(extra_string):
                    keycloak_realm = keycloak_realm[: -len(extra_string)]
                extra_string = "/"
                if keycloak_realm is not None and keycloak_realm.endswith(extra_string):
                    keycloak_realm = keycloak_realm[: -len(extra_string)]
                if (
                    keycloak_realm is not None
                    and "/" in keycloak_realm
                    or keycloak_realm is None
                ):
                    raise RuntimeError(
                        "Cannot extract the keycloak realm from the provided url."
                    )
            else:
                raise RuntimeError(
                    "Provided 'provider_discovery_uri' url is not a valid Keycloak metadata url, it does not contains /realms/."
                )
            keycloak_base_uri = f"{url.scheme}{url.netloc}{base_path}"

        if keycloak_base_uri is not None:
            self.keycloak_base_uri = keycloak_base_uri
        if self.keycloak_base_uri[-1] == "/":
            self.keycloak_base_uri = self.keycloak_base_uri[:-1]
        self.keycloak_realm = keycloak_realm
        provider_discovery_uri = (
            f"{self.keycloak_base_uri}/realms/{self.keycloak_realm}"
        )
        kwargs["provider_discovery_uri"] = provider_discovery_uri
        super().__init__(op_name=op_name, *args, **kwargs)

    @override
    def get_default_config(self) -> ProviderConfig:
        result = super().get_default_config()

        result["oidc_logout_query_string_redirect_parameter"] = "redirect_uri"
        return result



# type: ignore
import datetime

from django.contrib.auth import get_user_model
from django.contrib.auth.models import AbstractUser
from rest_framework import authentication, exceptions

from django_pyoidc.utils import OIDCCacheBackendForDjango
from django_pyoidc.views import OIDClient


class BaseOIDCAuthentication(authentication.BaseAuthentication):
    def __init__(self):
        # fixme : no multi-provider support here
        self.op_name = "default"
        self.general_cache_backend = OIDCCacheBackendForDjango(self.op_name)
        self.client = OIDClient(self.op_name)

    def authenticate(self, request):
        token = self.get_access_token(request)
        if token is None:
            return
        cache_key = self.general_cache_backend.generate_hashed_cache_key(token)
        try:
            access_token_claims = self.general_cache_backend["cache_key"]
        except KeyError:
            access_token_claims = self._introspect_access_token(token)
            if "active" not in access_token_claims:
                raise exceptions.AuthenticationFailed(
                    "Invalid identity provider reponse"
                )
            if not access_token_claims["active"]:
                raise exceptions.AuthenticationFailed("Account disabled")
            print(f"{access_token_claims=}")
            print(f"{token=}")
            # store it in cache
            current = datetime.datetime.now().strftime("%s")
            if "exp" not in access_token_claims:
                raise exceptions.AuthenticationFailed(
                    "No expiry set on the access token."
                )
            access_token_expiry = access_token_claims["exp"]
            exp = int(access_token_expiry) - int(current)
            self.general_cache_backend.set(cache_key, access_token_claims, exp)
        return self.get_user(token, access_token_claims), None

    def _introspect_access_token(self, token):
        request_args = {
            "token": token,
            "token_type_hint": "access_token",
        }
        client_auth_method = self.client.consumer.registration_response.get(
            "introspection_endpoint_auth_method", "client_secret_basic"
        )
        introspection = self.client.client_extension.do_token_introspection(
            request_args=request_args,
            authn_method=client_auth_method,
            endpoint=self.client.consumer.introspection_endpoint,
            # http_args={"headers" : {"content-type":"application/x-www-form-urlencoded"}}
        )
        print(f"{introspection=}")
        return introspection.to_dict()

    def get_access_token(self, request):
        # fixme : hardcoded header key in the following function call
        header = authentication.get_authorization_header(request)
        if not header:
            return None
        header = header.decode(authentication.HTTP_HEADER_ENCODING)
        print(f"{header=}")
        return header

    def get_user(self, token: str, access_token_claims) -> AbstractUser:
        raise NotImplementedError(
            f"Do no use {self.__class__.__name__} directly : inherit from it"
            f"and override get_user()"
        )


class DefaultOIDCAuthentication(BaseOIDCAuthentication):
    def get_user(self, token: str, access_token_claims) -> AbstractUser:
        User = get_user_model()
        user, _ = User.objects.get_or_create(id=token["sub"])
        return user
